/****************************************************************************
*
*  MiniSumo.c
*
*  Mini Sumo code for Wedgy
*
****************************************************************************/

#include <avr/io.h>

#include "Stinger.h"
#include "a2d.h"
#include "Avoid.h"
#include "Debug.h"
#include "DebugKey.h"
#include "Delay.h"
#include "Motor.h"
#include "Puts.h"
#include "Sensor.h"
#include "Uart.h"

#include <stdlib.h>

#define COUNTER_TURN_COUNT  200
#define COUNTER_ARC_COUNT   25    // 0 - 50 = arc, > 50 = SPIN

// When we randomize the turn count, we set gHuntTurnCount to a value between
// the min & max and set gHuntCount to the base.

#define AVOID_COUNT_BASE    ( COUNTER_ARC_COUNT )   // Base at Arc count so bot spins
#define AVOID_COUNT_MIN     ( AVOID_COUNT_BASE + 20 )
#define AVOID_COUNT_MAX     ( AVOID_COUNT_BASE + 100 )

// It takes about 6/10 of second to spin, so we'll ignore about 1/3 of a circle.

#define  COUNTER_EYE_RESET_COUNT ( COUNTER_ARC_COUNT + 25 )

#define  COUNTER_STRAIGHT_COUNT  400   // Start going straight

uns8  gIgnoreEye;
uns8  gEyeState;
uns8  gPrevEyeState;

uns8  gHuntDir;
uns16 gHuntCounter;

uns16 gHuntTurnCount = COUNTER_TURN_COUNT;

void RandomizeTurnCount( uns16 minCount, uns16 maxCount )
{
    uns32 randNum = rand();

#if 1
    randNum *= ( maxCount - minCount );
    randNum /= RAND_MAX;
    randNum += minCount;
#else
    randNum = maxCount;
#endif
    
    gHuntTurnCount = (uns16)randNum;
}

/***************************************************************************/
/**
*  Main program for the Mini Sumo
*/

int main( void )
{
    uns8  lineSide;
    uns8  i;
    uns8  sec;

    DBG( uns8  huntState; )
    DBG( uns8  prevHuntState; )

    // Initialization

    UART_Init();
    InitializeA2D();
    InitSensors();
    InitLED();

#if DEBUG
    UART_PutStr( "******************************\r\n" );
    UART_PutStr( "**                          **\r\n" );
    UART_PutStr( "** Stinger MiniSumo Program **\r\n" );
    UART_PutStr( "**                          **\r\n" );
    UART_PutStr( "******************************\r\n\r\n" );
#endif

    InitTimer();
    InitMotors();

    //puts( "Testing\n" );

    gDeferredNewline = 0;

    //gLineThresh = 12; // Special hack for Robothon

    DBG( puts( "LineThresh: " ); )
    DBG( putu8( gLineThresh ); )
    DBG( puts( "\n" ); )

    gHuntCounter = 0;

    gEyeState = 0;
    gPrevEyeState = 0;
    gIgnoreEye = 0;

    DBG( huntState = 0; )
    DBG( prevHuntState = 0; )

    switch ( PgmSelect() )
    {
        case 0: // Go Left
        {
            puts( "Initial Hunt Left\n" );
            gHuntDir = HUNT_LEFT;
            break;
        }
        
        case 1: // Go Straight
        {
            puts( "Initial Hunt Straight\n" );
            gHuntDir = HUNT_RIGHT;
            gHuntTurnCount = 0;
            gHuntCounter = 0;
            break;
        }
        
        case 2: // Go Right
        {
            puts( "Initial Hunt Right\n" );
            gHuntDir = HUNT_RIGHT;
            break;                        
        }
    }

    if ( IsBlackButtonPressed() )
    {
        // We want to wait for the black button to be released and re-pressed
        // to start the 5 second countdown.

        puts( "Waiting for Black Button to be released\n" );
        
        while ( 1 )
        {
            if ( !IsBlackButtonPressed() )
            {
                break;
            }

            BlueLED( 1 );
            GreenLED( 1 );

            for ( i = 0; i < 12; i++ ) 
            {
                if ( !IsBlackButtonPressed() )
                {
                    break;
                }
               
                WaitForTimer0Rollover();
            }

            BlueLED( 0 );
            GreenLED( 0 );
        
            for ( i = 0; i < 12; i++ ) 
            {
                if ( !IsBlackButtonPressed() )
                {
                    break;
                }

                WaitForTimer0Rollover();
            }
        }

        // Wait for 100 mSec debounce time
                
        for ( i = 0; i < 10; i++ ) 
        {
            WaitForTimer0Rollover();
        }
        
        puts( "Waiting for Black Button press to start 5 second delay\n" );

        while ( 1 )
        {
            if ( IsBlackButtonPressed() )
            {
                break;
            }

            BlueLED( 1 );
            GreenLED( 0 );

            for ( i = 0; i < 12; i++ ) 
            {
                if ( IsBlackButtonPressed() )
                {
                    break;
                }
               
                WaitForTimer0Rollover();
            }

            BlueLED( 0 );
            GreenLED( 1 );
        
            for ( i = 0; i < 12; i++ ) 
            {
                if ( IsBlackButtonPressed() )
                {
                    break;
                }

                WaitForTimer0Rollover();
            }
        }

        // OK, the black button has been pressed. Do the 5 second countdown.

        puts( "Starting 5 second countdown\n" );

        for ( sec = 0; sec < 5; sec++ ) 
        {
            BlueLED( 1 );
            GreenLED( 0 );

            for ( i = 0; i < 50; i++ ) 
            {
                WaitForTimer0Rollover();
            }
        
            BlueLED( 0 );
            GreenLED( 1 );
        
            for ( i = 0; i < 50; i++ ) 
            {
                WaitForTimer0Rollover();
            }
        }
    }
        
    InitLineThresh();

    // Turn on the eyes, so we can find our opponent.

    EnableLeftEye( TRUE );
    EnableRightEye( TRUE );

        puts( "Go\n" );

    while ( 1 )
    {
        // Wait for 100 Hz timer

        WaitForTimer0Rollover();
        gHuntCounter++;

#if DEBUG || 1
        if ( UART_IsCharAvailable() )
        { 
            UART_GetChar();
#if 1
            DebugKey();
#else
            /*
             * A key was pressed. pause and wait for another
             */

            puts( "Paused\n" );

            while ( !UART_IsCharAvailable() ) ;

            puts( "Go\n" );
            UART_GetChar();
#endif
        }
#endif

        ReadSensors();

#if USE_TILT
        if ( gTiltState != 0 )
        {
            AvoidTilt();
        }
#endif

#if DEBUG
        if (( gHuntCounter & 0x0F ) == 0 )
        {
            gDeferredNewline = 0;

            puts( "\rLL: " );
            putu8( gLineLO );
            puts( " " );
            putu8( gLineLI );
            puts( " LR: " );
            putu8( gLineRI );
            puts( " " );
            putu8( gLineRO );
            
            puts( " EL: " );
            putu8( gEyeL );
            puts( " ER: " );
            putu8( gEyeR );

#if 0
            puts( " IR: " );
            putu8( gIrDetected );
            puts( " (" );
            putu8( gDetectCount );
            puts( ")" );
#endif
            puts( " SL: " );
            putu8( gSpeedL );
            puts( " SR: " );
            putu8( gSpeedR );
#if USE_TILT
            puts( " TX: " );
            putu8( gTiltX );
            puts( " TY: " );
            putu8( gTiltY );
#endif
            puts( " C: " );
            putu16( gHuntCounter );                   

            gDeferredNewline = 1;
        } 
#endif

        gSpeedL = SPEED_FWD;
        gSpeedR = SPEED_FWD;

        //if ( !gIrDetected )
        {
            // Check line sensors

            lineSide = 0;

            if ( LineDetectedRight() )
            {
                DBG( puts( "LineR detected (" ); )
                DBG( putu8( gLineRI ); )
                DBG( puts( " " ); )
                DBG( putu8( gLineRO ); )
                DBG( puts( ") C:" ); )
                DBG( putu16( gHuntCounter ); )
                DBG( puts( "\n" ); )
                
                lineSide |= LINE_DETECTED_RIGHT;
                //if ( !gIrDetected )
                {
                    // Set the hunt direction to be left so that we hunt "away" from the line.
    
                    gHuntDir = HUNT_LEFT;
                }
            }
            if ( LineDetectedLeft() )
            {
                DBG( puts( "LineL detected (" ); )
                DBG( putu8( gLineLO  ); )
                DBG( puts( " " ); )
                DBG( putu8( gLineLI  ); )
                DBG( puts( ") C:" ); )
                DBG( putu16( gHuntCounter ); )
                DBG( puts( "\n" ); )
                
                lineSide |= LINE_DETECTED_LEFT;
                //if ( !gIrDetected )
                {
                   // Set the hunt direction to be right so that we hunt "away" from the line.
                
                   gHuntDir = HUNT_RIGHT;
                }
            }
                
            if ( lineSide != 0 )
            {
                AvoidEdge( lineSide );  
                    
                // We set the Hunt Counter to 50 so that it will spin as opposed
                // to arc.
                
                gHuntCounter = AVOID_COUNT_BASE;
                
                RandomizeTurnCount( AVOID_COUNT_MIN, AVOID_COUNT_MAX );
                
                // Raise the eye threshold since we hit the line. Otherwise, if
                // we've locked on to something outside the ring we'll just
                // keep running towards it.
                
                gIgnoreEye = 1;
            }
        }

      // Bigger number = closer distance

      gEyeState = 0;

#if 1
      if ( !gIgnoreEye )
      {
          if ( EyeDetectedLeft() )
          {
             // We have some object in the left

             gEyeState |= EYE_DETECTED_LEFT;
          }
          if ( EyeDetectedRight() )
          {
             // We have some object in the right

             gEyeState |= EYE_DETECTED_RIGHT;
          }
      }
#endif

      if ( gEyeState == EYE_DETECTED_LEFT )
      {
         // Object detected on the left side only. Turn towards it

         gSpeedL = SPEED_FWD_MED;
         
         //if ( gHuntDir != HUNT_LEFT )
         {
            //DBG( puts( "Resetting gHuntCounter L\n" ); )
            gHuntCounter = 0;
            gHuntTurnCount = COUNTER_TURN_COUNT;
         }
         gHuntDir = HUNT_LEFT;
      }
      else
      if ( gEyeState == EYE_DETECTED_RIGHT )
      {
         // Object detected on the right side only. Turn towards it

         gSpeedR = SPEED_FWD_MED;
         
         //if ( gHuntDir != HUNT_RIGHT )
         {
            //DBG( puts( "Resetting gHuntCounter R\n" ); )
            gHuntCounter = 0;
            gHuntTurnCount = COUNTER_TURN_COUNT;
         }
         gHuntDir = HUNT_RIGHT;
      }
      else      
      //if (( gEyeState == 0 ) && !gIrDetected )
      if ( gEyeState == 0 )
      {
         // No object detected and the other guy isn't close, so go into
         // hunt mode.

         if ( gHuntCounter < gHuntTurnCount )
         {
            if ( gHuntDir == HUNT_LEFT )
            {
               DBG( huntState = HUNT_STATE_SPIN_LEFT; )

               // Spin Left

               if ( gHuntCounter < COUNTER_ARC_COUNT )
               {
                  // Arc forward turning left

                  gSpeedL = SPEED_FWD_MED;
               }
               else
               {
                  // Spin to the left

                  gSpeedL = SPEED_BWD;
               }
               gSpeedR = SPEED_FWD;
            }
            else
            {
               DBG( huntState = HUNT_STATE_SPIN_RIGHT; )

               // Spin Right

               gSpeedL = SPEED_FWD;
               if ( gHuntCounter < COUNTER_ARC_COUNT )
               {
                  // Arc forward turning right

                  gSpeedR = SPEED_FWD_MED;
               }
               else
               {
                  // Spin right

                  gSpeedR = SPEED_BWD;
               }
            }
         }
         else
         if ( gHuntCounter < COUNTER_STRAIGHT_COUNT )
         {
            DBG( huntState = HUNT_STATE_GO_STRAIGHT; )

            // Go forward

            gSpeedL = SPEED_FWD;
            gSpeedR = SPEED_FWD;
         }
         else
         {
            // Counter wrapped and we still didn't find anybody.
            // Swap hunt direction so we don't get "stuck in a corner"

            gHuntCounter = 0;
            gHuntDir = SWAP_HUNT_DIR( gHuntDir );
         }

         if ( gHuntCounter >= COUNTER_EYE_RESET_COUNT )
         {
            gIgnoreEye = 0;
         }
      }
#if DEBUG
      if ( gEyeState != gPrevEyeState )
      {
         puts( "EyeState changed: " );
         if (( gEyeState & EYE_DETECTED_LEFT ) != 0 )
         {
            putc( 'L' );
         }
         else
         {
            putc( ' ' );
         }
         if (( gEyeState & EYE_DETECTED_RIGHT ) != 0 )
         {
            putc( 'R' );
         }
         else
         {
            putc( ' ' );
         }
         puts( " EyeL:" );
         putu8( gEyeL );
         puts( " EyeR:" );
         putu8( gEyeR );
         puts( " C: " );
         putu16( gHuntCounter );
         puts( "\n" );
   
         gPrevEyeState = gEyeState;
      }
      if ( huntState != prevHuntState )
      {
         if ( huntState != 0 )
         {
            puts( "Hunt: " );

            switch ( huntState )
            {
               case HUNT_STATE_SPIN_LEFT:
               {
                  puts( "Spin Left" );
                  break;
               }

               case HUNT_STATE_SPIN_RIGHT:
               {
                  puts( "Spin Right" );
                  break;
               }

               case HUNT_STATE_GO_STRAIGHT:
               {
                  puts( "Go Forward" );
                  break;
               }

               default:
               {
                  puts( "*** Unknown ***" );
                  break;
               }
            }

            puts( " C: ");
            putu16( gHuntCounter );
            puts( "\n" );
         }
         prevHuntState = huntState;
      }
#endif
      
      PulseMotors();
   }

   return 0;

} // main

/** @} */

